from typing import Optional, Callable

import panel
from atap_corpus.corpus.corpus import DataFrameCorpus
from panel.viewable import Viewer

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view import ViewWrapperWidget, NotifierService

panel.extension(notifications=True)


class CorpusLoader(Viewer):
    """
    Public interface for the CorpusLoader module. Maintains a reference to the logic Controller and the GUI wrapper.
    A CorpusLoader object can be used as a Panel component, i.e. will render in a Panel GUI.
    The build_callback_fn will be called when a corpus is built (can be set using set_build_callback()).
    """

    def __init__(self, root_directory: str, **params):
        super().__init__(**params)
        self.controller: Controller = Controller(NotifierService(), root_directory)
        self.view: ViewWrapperWidget = ViewWrapperWidget(self.controller)

    def __panel__(self):
        return self.view

    def set_build_callback(self, callback: Callable, *args, **kwargs):
        """
        Allows a callback function to be set when a corpus has completed building
        :param callback: the function to call when a corpus has been built
        :param args: positional arguments to pass onto the callback function
        :param kwargs: keyword arguments to pass onto the callback function
        """
        self.controller.set_build_callback(callback, *args, **kwargs)

    def get_corpus(self) -> Optional[DataFrameCorpus]:
        """
        :return: the last DataFrameCorpus object that was built. If none have been built, returns None.
        """
        return self.controller.get_latest_corpus()

    def get_corpora(self) -> list[DataFrameCorpus]:
        return self.controller.get_corpora()



from glob import glob
from io import BytesIO
from os.path import join, isdir, basename, dirname
from typing import Optional, Callable

from atap_corpus.corpus.corpora import UniqueCorpora
from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame

from atap_corpus_loader.controller.CorpusExportService import CorpusExportService
from atap_corpus_loader.controller.FileLoaderService import FileLoaderService, FileLoadError
from atap_corpus_loader.controller.OniAPIService import OniAPIService
from atap_corpus_loader.controller.data_objects import FileReference, ZipFileReference, ViewCorpusInfo
from atap_corpus_loader.controller.data_objects.CorpusHeader import CorpusHeader
from atap_corpus_loader.controller.data_objects.DataType import DataType
from atap_corpus_loader.controller.file_loader_strategy.FileLoaderFactory import ValidFileType
from atap_corpus_loader.view.notifications import NotifierService


class Controller:
    """
    Provides methods for indirection between the corpus loading logic and the user interface
    Holds a reference to the latest corpus built.
    The build_callback_fn will be called when a corpus is built (can be set using set_build_callback()).
    """
    def __init__(self, notifier_service: NotifierService, root_directory: str):
        self.root_directory: str = root_directory

        self.file_loader_service: FileLoaderService = FileLoaderService()
        self.oni_api_service: OniAPIService = OniAPIService()
        self.corpus_export_service: CorpusExportService = CorpusExportService()
        self.notifier_service: NotifierService = notifier_service

        self.text_header: Optional[CorpusHeader] = None
        self.corpus_link_header: Optional[CorpusHeader] = None
        self.meta_link_header: Optional[CorpusHeader] = None

        self.corpus_headers: list[CorpusHeader] = []
        self.meta_headers: list[CorpusHeader] = []

        self.corpora: UniqueCorpora = UniqueCorpora()
        self.latest_corpus: Optional[DataFrameCorpus] = None

        self.build_callback_fn: Optional[Callable] = None
        self.build_callback_args: list = []
        self.build_callback_kwargs: dict = {}

    def display_error(self, error_msg: str):
        self.notifier_service.notify_error(error_msg)

    def display_success(self, success_msg: str):
        self.notifier_service.notify_success(success_msg)

    def set_build_callback(self, callback: Callable, *args, **kwargs):
        if not callable(callback):
            raise ValueError("Provided callback function must be a callable")
        self.build_callback_fn = callback
        self.build_callback_args = args
        self.build_callback_kwargs = kwargs

    def get_corpora(self) -> list[DataFrameCorpus]:
        return self.corpora.items()

    def get_latest_corpus(self) -> Optional[DataFrameCorpus]:
        return self.latest_corpus

    def get_loaded_corpus_df(self) -> Optional[DataFrame]:
        if self.latest_corpus is None:
            return None
        return self.latest_corpus.to_dataframe()

    def load_corpus_from_filepaths(self, filepath_ls: list[FileReference]) -> bool:
        for filepath in filepath_ls:
            try:
                self.file_loader_service.add_corpus_filepath(filepath)
                self.corpus_headers = self.file_loader_service.get_inferred_corpus_headers()
            except FileLoadError as e:
                self.file_loader_service.remove_corpus_filepath(filepath)
                self.display_error(str(e))
                return False

        self.display_success("Corpus files loaded successfully")
        return True

    def load_meta_from_filepaths(self, filepath_ls: list[FileReference]) -> bool:
        for filepath in filepath_ls:
            try:
                self.file_loader_service.add_meta_filepath(filepath)
                self.meta_headers = self.file_loader_service.get_inferred_meta_headers()
            except FileLoadError as e:
                self.file_loader_service.remove_meta_filepath(filepath)
                self.display_error(str(e))
                return False

        self.display_success("Metadata files loaded successfully")
        return True

    def build_corpus(self, corpus_name: str) -> bool:
        if self.is_meta_added():
            if (self.corpus_link_header is None) or (self.meta_link_header is None):
                self.display_error("Cannot build without link headers set. Select a corpus header and a meta header as linking headers in the dropdowns")
                return False

        try:
            self.latest_corpus = self.file_loader_service.build_corpus(corpus_name, self.corpus_headers,
                                                                       self.meta_headers, self.text_header,
                                                                       self.corpus_link_header, self.meta_link_header)
        except FileLoadError as e:
            self.display_error(str(e))
            return False
        except Exception as e:
            self.display_error(f"Unexpected error building corpus: {e}")
            return False

        try:
            if self.build_callback_fn is not None:
                self.build_callback_fn(*self.build_callback_args, **self.build_callback_kwargs)
        except Exception as e:
            self.display_error(f"Build callback error: {e}")
            return False

        self.display_success(f"Corpus {self.latest_corpus.name} built successfully")
        self.corpora.add(self.latest_corpus)

        return True

    def get_corpora_info(self) -> list[ViewCorpusInfo]:
        corpora_info: list[ViewCorpusInfo] = []

        for corpus in self.corpora.items():
            corpus_as_df: DataFrame = corpus.to_dataframe()

            corpus_id: str = str(corpus.id)
            name: Optional[str] = corpus.name
            num_rows: int = len(corpus)
            headers: list[str] = []
            dtypes: list[str] = []
            dtype: str
            for header_name, dtype_obj in corpus_as_df.dtypes.items():
                try:
                    dtype = DataType(str(dtype_obj).lower()).name
                except KeyError:
                    dtype = str(dtype_obj).upper()
                dtypes.append(dtype)
                headers.append(str(header_name))

            corpora_info.append(ViewCorpusInfo(corpus_id, name, num_rows, headers, dtypes))

        return corpora_info

    def delete_corpus(self, corpus_id: str):
        self.corpora.remove(corpus_id)

    def rename_corpus(self, corpus_id: str, new_name: str):
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_id)
        if corpus is None:
            self.display_error(f"No corpus with id {corpus_id} found")
            return
        if corpus.name == new_name:
            return

        try:
            corpus.name = new_name
        except ValueError as e:
            self.display_error(str(e))

    def get_loaded_file_counts(self) -> dict[str, int]:
        corpus_file_set = set(self.file_loader_service.get_loaded_corpus_files())
        meta_file_set = set(self.file_loader_service.get_loaded_meta_files())
        file_set = corpus_file_set | meta_file_set

        file_counts: dict[str, int] = {"Total files": len(file_set)}
        for file_ref in file_set:
            extension = file_ref.get_extension().upper()
            if file_counts.get(extension) is None:
                file_counts[extension] = 1
            else:
                file_counts[extension] += 1

        return file_counts

    def unload_filepaths(self, filepath_ls: list[FileReference]):
        for filepath in filepath_ls:
            self.file_loader_service.remove_meta_filepath(filepath)
            self.file_loader_service.remove_corpus_filepath(filepath)

        if len(self.file_loader_service.get_loaded_corpus_files()) == 0:
            self.text_header = None
            self.corpus_headers = []
            self.corpus_link_header = None
        if len(self.file_loader_service.get_loaded_meta_files()) == 0:
            self.meta_headers = []
            self.meta_link_header = None

    def unload_all(self):
        self.file_loader_service.remove_all_files()

        self.text_header = None
        self.corpus_headers = []
        self.meta_headers = []
        self.corpus_link_header = None
        self.meta_link_header = None

    def get_loaded_corpus_files(self) -> list[FileReference]:
        return self.file_loader_service.get_loaded_corpus_files()

    def get_loaded_meta_files(self) -> list[FileReference]:
        return self.file_loader_service.get_loaded_meta_files()

    def get_corpus_headers(self) -> list[CorpusHeader]:
        return self.corpus_headers

    def get_meta_headers(self) -> list[CorpusHeader]:
        return self.meta_headers

    def get_inferred_corpus_headers(self) -> list[CorpusHeader]:
        return self.file_loader_service.get_inferred_corpus_headers()

    def get_inferred_meta_headers(self) -> list[CorpusHeader]:
        return self.file_loader_service.get_inferred_meta_headers()

    def get_text_header(self) -> Optional[CorpusHeader]:
        return self.text_header

    def get_corpus_link_header(self) -> Optional[CorpusHeader]:
        return self.corpus_link_header

    def get_meta_link_header(self) -> Optional[CorpusHeader]:
        return self.meta_link_header

    def get_all_datatypes(self) -> list[str]:
        return [d.name for d in DataType]

    def get_valid_filetypes(self) -> list[str]:
        return [ft.name for ft in ValidFileType]

    def is_corpus_added(self) -> bool:
        return len(self.corpus_headers) > 0

    def is_meta_added(self) -> bool:
        return len(self.meta_headers) > 0

    def get_loaded_file_count(self) -> int:
        corpus_file_set = set(self.file_loader_service.get_loaded_corpus_files())
        meta_file_set = set(self.file_loader_service.get_loaded_meta_files())
        file_set = corpus_file_set | meta_file_set

        return len(file_set)

    def update_corpus_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, corpus_header in enumerate(self.corpus_headers):
            if header == corpus_header:
                self.corpus_headers[i] = header

    def update_meta_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, meta_header in enumerate(self.meta_headers):
            if header == meta_header:
                self.meta_headers[i] = header

    def set_text_header(self, text_header: Optional[str]):
        if text_header is None:
            self.text_header = None
            return

        for header in self.corpus_headers:
            if header.name == text_header:
                self.text_header = header
                header.datatype = DataType.TEXT
                header.include = True
                return

    def set_corpus_link_header(self, link_header_name: Optional[str]):
        for header in self.corpus_headers:
            if header.name == link_header_name:
                self.corpus_link_header = header
                header.include = True
                return
        self.corpus_link_header = None

    def set_meta_link_header(self, link_header_name: Optional[str]):
        for header in self.meta_headers:
            if header.name == link_header_name:
                self.meta_link_header = header
                header.include = True
                return
        self.meta_link_header = None

    def retrieve_all_files(self) -> list[FileReference]:
        all_relative_paths: list[str] = glob("**", root_dir=self.root_directory, recursive=True)
        all_full_paths: list[str] = []
        for path in all_relative_paths:
            full_path = join(self.root_directory, path)
            if not isdir(full_path):
                all_full_paths.append(join(self.root_directory, path))

        all_file_refs: list[FileReference] = []
        for idx, full_path in enumerate(all_full_paths):
            if full_path.endswith('.zip'):
                zip_file_refs: list[FileReference] = ZipFileReference.get_zip_internal_file_refs(self.root_directory, full_path)
                all_file_refs.extend(zip_file_refs)
            else:
                file_ref = FileReference(self.root_directory, dirname(full_path), basename(full_path))
                all_file_refs.append(file_ref)

        all_file_refs.sort(key=lambda ref: ref.get_full_path())

        return all_file_refs

    def get_export_types(self) -> list[str]:
        return self.corpus_export_service.get_filetypes()

    def export_corpus(self, corpus_id: str, filetype: str) -> Optional[BytesIO]:
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_id)
        if corpus is None:
            self.display_error(f"No corpus with id {corpus_id} found")
            return None

        try:
            return self.corpus_export_service.export(corpus.to_dataframe(), filetype)
        except ValueError as e:
            self.display_error(str(e))



from panel import Tabs

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.view.gui import AbstractWidget, FileLoaderWidget, OniLoaderWidget, CorpusInfoWidget


class ViewWrapperWidget(AbstractWidget):
    """
    A wrapper class that holds different loading method interfaces within a Tab
    """
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller: Controller = controller

        self.file_loader: FileLoaderWidget = FileLoaderWidget(self, controller)
        self.oni_loader: OniLoaderWidget = OniLoaderWidget(controller)
        self.corpus_display: CorpusInfoWidget = CorpusInfoWidget(controller)
        self.display_idx: int = 2

        self.panel = Tabs(("File Loader", self.file_loader),
                          ("Oni Loader", self.oni_loader),
                          ("Corpus Overview", self.corpus_display))
        self.children = [self.file_loader, self.oni_loader, self.corpus_display]

    def update_display(self):
        pass

    def load_corpus_from_filepaths(self, filepath_ls: list[FileReference]) -> bool:
        success = self.controller.load_corpus_from_filepaths(filepath_ls)
        self.update_displays()
        return success

    def load_meta_from_filepaths(self, filepath_ls: list[FileReference]) -> bool:
        success = self.controller.load_meta_from_filepaths(filepath_ls)
        self.update_displays()
        return success

    def build_corpus(self, corpus_name: str):
        success: bool = self.controller.build_corpus(corpus_name)
        if success:
            self.update_displays()
            self.panel.active = self.display_idx
            self.file_loader.unload_all()



from os.path import join, relpath, dirname, basename
from tempfile import NamedTemporaryFile
from zipfile import ZipFile


class FileReference:
    """
    A general purpose object to hold information regarding a specific file in the file system.
    Folder structure is preserved as a path-like string
    """
    def __init__(self, root_directory: str, directory_path: str, filename: str):
        """
        :param root_directory: the directory from which all files are relative to. This is specified in the CorpusLoader constructor
        :param directory_path: the path to the parent directory of the file. This is not relative to the root_directory
        :param filename: the name and extension of the file, e.g. test.txt
        """
        self.root_directory: str = root_directory
        self.directory_path: str = directory_path
        self.filename: str = filename

    def __eq__(self, other):
        if not isinstance(other, FileReference):
            return False
        return self.get_full_path() == other.get_full_path()

    def __hash__(self):
        return hash(self.get_full_path())

    def __str__(self):
        return self.get_full_path()

    def __repr__(self):
        return self.get_full_path()

    def resolve_real_file_path(self) -> str:
        """
        Provides a real addressable path to the file contents. If the FileReference object is an instance of
        ZipFileReference, the file is extracted, placed in a temporary file, and the temporary file path will be provided
        :return: the full addressable path of the file
        """
        return self.get_full_path()

    def get_full_path(self) -> str:
        """
        :return: the joined directory_path and filename to form the full path of the file
        """
        return join(self.directory_path, self.filename)

    def get_directory_path(self):
        return self.directory_path

    def get_filename(self) -> str:
        return self.filename

    def get_extension(self) -> str:
        if '.' not in self.filename:
            return ''
        return self.filename.split('.')[-1]

    def get_relative_path(self):
        """
        :return: the full path of the file relative to the root_directory
        """
        return relpath(self.get_full_path(), self.root_directory)

    def is_zipped(self) -> bool:
        """
        If True, the file is contained within a zip archive. In this case, the path returned by get_full_path()
        is not a real addressable path, just a string representation of where the file is located. A real addressable
        path can be obtained from resolve_real_file_path()
        :return: True if FileReference object is an instance of ZipFileReference, False otherwise
        """
        return False


class ZipFileReference(FileReference):
    @staticmethod
    def get_zip_internal_file_refs(root_directory: str, zip_file_path: str) -> list[FileReference]:
        """
        Accepts a zip file and the root_directory for files and provides a list of FileReference objects that correspond
        to the zipped files within the zip archive.
        :param root_directory: the directory from which all files are relative to. This is specified in the CorpusLoader constructor
        :param zip_file_path: the path to the zip archive that holds the files to be listed.
        :return: a list of FileReference objects corresponding to the files within the zip archive
        """
        with ZipFile(zip_file_path) as zip_f:
            info_list = zip_f.infolist()

        file_refs: list[FileReference] = []
        for info in info_list:
            if info.is_dir():
                continue
            internal_directory: str = dirname(info.filename)
            filename: str = basename(info.filename)
            zip_ref: FileReference = ZipFileReference(root_directory, zip_file_path, internal_directory, filename)
            file_refs.append(zip_ref)

        return file_refs

    def __init__(self, root_directory: str, zip_file_path: str, internal_directory: str, filename: str):
        """
        :param root_directory: the directory from which all files are relative to. This is specified in the CorpusLoader constructor
        :param zip_file_path: the path to the zip file that holds this zipped file. This is not relative to the root_directory
        :param internal_directory: the path within the zip file to this zipped file
        :param filename: the name and extension of the file, e.g. test.txt
        """
        super().__init__(root_directory, zip_file_path, filename)
        self.internal_directory: str = internal_directory

        self.zip_file = None

    def get_full_path(self) -> str:
        """
        :return: the joined directory_path, internal_directory, and filename to form the full addressable path of the file
        """
        return join(self.directory_path, self.internal_directory, self.filename)

    def is_zipped(self) -> bool:
        """
        If True, the file is contained within a zip archive. In this case, the path returned by get_full_path()
        is not a real addressable path, just a string representation of where the file is located. A real addressable
        path can be obtained from resolve_real_file_path()
        :return: True as FileReference object is an instance of ZipFileReference
        """
        return True

    def resolve_real_file_path(self) -> str:
        """
        Provides a real addressable path to the file contents. The zipped file is extracted,
        placed in a temporary file, and the temporary file path is provided
        :return: the full addressable path of the file
        """
        internal_path = join(self.internal_directory, self.filename)
        zip_file = ZipFile(self.directory_path)
        with zip_file.open(internal_path, force_zip64=True) as zip_f:
            file_content = zip_f.read()
        with NamedTemporaryFile(delete=False) as temp_f:
            temp_f.write(file_content)
            filepath = temp_f.name

        return filepath



import os
from typing import Optional

from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame, merge, concat

from atap_corpus_loader.controller.data_objects import FileReference, CorpusHeader
from atap_corpus_loader.controller.file_loader_strategy import FileLoaderStrategy, FileLoaderFactory, FileLoadError


class FileLoaderService:
    """
    Provides methods that handle the logic of loading files and building the DataFrameCorpus object from the loaded
    files.
    Maintains a reference to files loaded as corpus files and files loaded as metadata files.
    """
    def __init__(self):
        self.corpus_file_refs: list[FileReference] = []
        self.meta_file_refs: list[FileReference] = []

    def get_loaded_corpus_files(self) -> list[FileReference]:
        return self.corpus_file_refs.copy()

    def get_loaded_meta_files(self) -> list[FileReference]:
        return self.meta_file_refs.copy()

    def add_corpus_filepath(self, corpus_filepath: FileReference):
        if corpus_filepath in self.corpus_file_refs:
            return

        FileLoaderService._check_filepath_permissions(corpus_filepath)
        self.corpus_file_refs.append(corpus_filepath)

    def add_meta_filepath(self, meta_filepath: FileReference):
        if meta_filepath in self.meta_file_refs:
            return

        FileLoaderService._check_filepath_permissions(meta_filepath)
        self.meta_file_refs.append(meta_filepath)

    def remove_corpus_filepath(self, corpus_filepath: FileReference):
        if corpus_filepath in self.corpus_file_refs:
            self.corpus_file_refs.remove(corpus_filepath)

    def remove_meta_filepath(self, meta_filepath: FileReference):
        if meta_filepath in self.meta_file_refs:
            self.meta_file_refs.remove(meta_filepath)

    def remove_all_files(self):
        self.corpus_file_refs = []
        self.meta_file_refs = []

    def get_inferred_corpus_headers(self) -> list[CorpusHeader]:
        return FileLoaderService._get_file_headers(self.corpus_file_refs)

    def get_inferred_meta_headers(self) -> list[CorpusHeader]:
        return FileLoaderService._get_file_headers(self.meta_file_refs)

    def build_corpus(self, corpus_name: str,
                     corpus_headers: list[CorpusHeader],
                     meta_headers: list[CorpusHeader],
                     text_header: CorpusHeader,
                     corpus_link_header: Optional[CorpusHeader],
                     meta_link_header: Optional[CorpusHeader]) -> DataFrameCorpus:
        corpus_df: DataFrame = FileLoaderService._get_concatenated_dataframe(self.corpus_file_refs, corpus_headers)
        meta_df: DataFrame = FileLoaderService._get_concatenated_dataframe(self.meta_file_refs, meta_headers)

        load_corpus: bool = len(corpus_headers) > 0
        load_meta: bool = len(meta_headers) > 0

        final_df: DataFrame
        if load_corpus and load_meta:
            final_df = merge(left=corpus_df, right=meta_df,
                             left_on=corpus_link_header.name, right_on=meta_link_header.name,
                             how='inner', suffixes=(None, '_meta'))
        elif load_corpus:
            final_df = corpus_df
        elif load_meta:
            final_df = meta_df
        else:
            raise ValueError("No corpus headers or metadata headers provided")

        return DataFrameCorpus.from_dataframe(df=final_df, col_doc=text_header.name, name=corpus_name)

    @staticmethod
    def _check_filepath_permissions(file_ref: FileReference):
        filepath: str
        if file_ref.is_zipped():
            filepath = file_ref.get_directory_path()
        else:
            filepath = file_ref.get_full_path()
        if not os.path.exists(filepath):
            raise FileLoadError(f"No file found at: {filepath}")
        if not os.access(filepath, os.R_OK):
            raise FileLoadError(f"No permissions to read the file at: {filepath}")

    @staticmethod
    def _get_file_headers(file_refs: list[FileReference]) -> list[CorpusHeader]:
        headers: Optional[list[CorpusHeader]] = None
        for ref in file_refs:
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_headers: list[CorpusHeader] = file_loader.get_inferred_headers()
            except UnicodeDecodeError:
                raise FileLoadError(f"Error loading file at {ref.get_relative_path()}: file is not UTF-8 encoded")
            except Exception as e:
                raise FileLoadError(f"Error loading file at {ref.get_relative_path()}: {e}")

            if headers is None:
                headers = path_headers
            elif set(headers) != set(path_headers):
                raise FileLoadError(f"Incompatible headers within loaded files")

        if headers is None:
            headers = []

        return headers

    @staticmethod
    def _get_concatenated_dataframe(file_refs: list[FileReference], headers: list[CorpusHeader]) -> DataFrame:
        if len(file_refs) == 0:
            return DataFrame()
        df_list: list[DataFrame] = []
        for ref in file_refs:
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_df: DataFrame = file_loader.get_dataframe(headers)
            except UnicodeDecodeError:
                raise FileLoadError(f"Error loading file at {ref.get_relative_path()}: file is not UTF-8 encoded")
            except Exception as e:
                raise FileLoadError(f"Error loading file at {ref.get_relative_path()}: {e}")

            df_list.append(path_df)

        return concat(df_list, ignore_index=True)



from fnmatch import fnmatch

import panel
from panel import Row, Column
from panel.widgets import Button, MultiSelect, TextInput, Select, Checkbox

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.view.gui import AbstractWidget


class FileSelectorWidget(AbstractWidget):
    def __init__(self, view_handler: AbstractWidget, controller: Controller):
        super().__init__()
        self.view_handler: AbstractWidget = view_handler
        self.controller: Controller = controller

        self.select_all_button = Button(name="Select all", width=95,
                                        button_style="solid", button_type="primary")
        self.select_all_button.on_click(self.select_all)

        self.filter_input = TextInput(placeholder="Filter displayed files (supports wildcard syntax)",
                                      sizing_mode='stretch_width')
        self.filter_input.param.watch(self._on_filter_change, ['value_input'])

        self.show_hidden_files_checkbox = Checkbox(name="Show hidden", value=False, align="center")
        self.show_hidden_files_checkbox.param.watch(self._on_filter_change, ['value'])

        self.file_type_filter = Select(width=150)
        self.file_type_filter.options = ['All valid filetypes'] + self.controller.get_valid_filetypes()
        self.file_type_filter.param.watch(self._on_filter_change, ['value'])

        self.selector_widget = MultiSelect(size=20, sizing_mode='stretch_width')

        self.panel = Column(
            Row(self.select_all_button),
            Row(self.filter_input,
                self.show_hidden_files_checkbox,
                self.file_type_filter),
            Row(self.selector_widget),
            width=700)

        panel.state.add_periodic_callback(self.update_display, period=1000)
        self.update_display()
        self.select_all()

    def update_display(self):
        loaded_corpus_files: list[FileReference] = self.controller.get_loaded_corpus_files()
        loaded_meta_files: list[FileReference] = self.controller.get_loaded_meta_files()

        all_file_refs: list[FileReference] = self.controller.retrieve_all_files()
        filtered_refs: list[FileReference] = self._filter_file_refs(all_file_refs)

        old_selected_refs: list[FileReference] = self.selector_widget.value.copy()
        filtered_selected_refs: list[FileReference] = [f for f in old_selected_refs if f in filtered_refs]

        filtered_files_dict: dict[str, FileReference] = {}
        checkmark_symbol = "\U00002714"
        for ref in filtered_refs:
            file_repr = ref.get_relative_path()
            if ref in loaded_corpus_files:
                file_repr += f" {checkmark_symbol} [corpus]"
            if ref in loaded_meta_files:
                file_repr += f" {checkmark_symbol} [meta]"
            filtered_files_dict[file_repr] = ref

        self.selector_widget.options = filtered_files_dict
        self.selector_widget.value = filtered_selected_refs

    def _filter_file_refs(self, file_refs: list[FileReference]) -> list[FileReference]:
        filtered_refs: list[FileReference] = []

        valid_file_types: list[str] = self.controller.get_valid_filetypes()
        selected_file_types: list[str]
        if self.file_type_filter.value in valid_file_types:
            selected_file_types = [self.file_type_filter.value.upper()]
        else:
            selected_file_types = [ft.upper() for ft in valid_file_types]

        filter_str = f"*{self.filter_input.value_input}*"
        for ref in file_refs:
            if not fnmatch(ref.get_relative_path(), filter_str):
                continue
            if ref.get_extension().upper() not in selected_file_types:
                continue
            is_hidden_file: bool = ref.get_filename().startswith('.')
            if is_hidden_file and not self.show_hidden_files_checkbox.value:
                continue

            filtered_refs.append(ref)

        return filtered_refs

    def _on_filter_change(self, *_):
        self.update_display()
        self.select_all()

    def select_all(self, *_):
        self.selector_widget.value = list(self.selector_widget.options.values())

    def get_selector_value(self) -> list[FileReference]:
        return self.selector_widget.value



from abc import ABC, abstractmethod
from typing import Callable

from panel import Row, Column
from panel.layout import Panel
from panel.widgets import Button


class AbstractWidget(ABC):
    """
    An abstract class for Panel GUI widgets. Provides methods to set widget visibility and update the display of all
    child widgets.
    """
    def __init__(self):
        self.panel: Panel = Row()
        self.children: list[AbstractWidget] = []

    def __panel__(self) -> Panel:
        return self.panel

    def update_displays(self):
        self.update_display()
        for child in self.children:
            child.update_displays()

    def get_visibility(self) -> bool:
        return self.panel.visible

    def set_visibility(self, is_visible: bool):
        self.panel.visible = is_visible

    def toggle_visibility(self):
        self.panel.visible = not self.panel.visible

    def create_confirmation_box(self, *_, confirm_callable: Callable):
        def confirm_action(_):
            confirm_callable()
            response.clear()

        def cancel_action(_):
            response.clear()

        confirmation = Button(name='Confirm')
        confirmation.on_click(confirm_action)
        cancel = Button(name='Cancel')
        cancel.on_click(cancel_action)

        response = Column("Are you sure?", Row(confirmation, cancel))
        response.show()

    @abstractmethod
    def update_display(self):
        raise NotImplementedError()



from panel import Row, Spacer, Column, HSpacer
from panel.pane import Markdown
from panel.widgets import Button, TextInput

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.view import ViewWrapperWidget
from atap_corpus_loader.view.gui import AbstractWidget
from atap_corpus_loader.view.gui.FileSelectorWidget import FileSelectorWidget
from atap_corpus_loader.view.gui.MetaEditorWidget import MetaEditorWidget


class FileLoaderWidget(AbstractWidget):
    def __init__(self, view_handler: ViewWrapperWidget, controller: Controller):
        super().__init__()
        self.view_handler: ViewWrapperWidget = view_handler
        self.controller: Controller = controller
        
        self.load_as_corpus_button: Button = Button(name='Load as corpus', width=130,
                                                    button_style='outline', button_type='success')
        self.load_as_corpus_button.on_click(self.load_as_corpus)
        self.load_as_meta_button: Button = Button(name='Load as metadata', width=130,
                                                  button_style='outline', button_type='success')
        self.load_as_meta_button.on_click(self.load_as_meta)

        self.unload_all_button: Button = Button(name="Unload all", width=100, button_style='solid',
                                                button_type='danger', disabled=True)
        self.unload_all_button.on_click(self.unload_all)
        self.unload_selected_button: Button = Button(name="Unload selected", width=100, button_style='outline',
                                                     button_type='danger', disabled=True)
        self.unload_selected_button.on_click(self.unload_selected)

        self.loaded_file_info = Markdown()
        
        self.corpus_name_input = TextInput(placeholder='Corpus name', width=150, visible=False)
        self.build_button = Button(name='Build corpus', button_style='solid', button_type='success', visible=False)
        self.build_button.on_click(self.build_corpus)

        self.file_selector = FileSelectorWidget(view_handler, controller)
        self.meta_editor = MetaEditorWidget(view_handler, controller)

        self.panel = Row(
            Column(
                self.file_selector,
                Row(Column(
                    Row(self.load_as_corpus_button,
                        self.load_as_meta_button),
                    Row(self.corpus_name_input,
                        self.build_button)
                ),
                    self.loaded_file_info,
                    HSpacer(),
                    self.unload_selected_button,
                    self.unload_all_button,
                    width=700)
            ),
            Spacer(width=50),
            self.meta_editor)
        self.children = [self.file_selector, self.meta_editor]
        self.update_display()

    def update_display(self):
        files_added: bool = self.controller.is_meta_added() or self.controller.is_corpus_added()
        self._set_build_buttons_status(files_added)
        self.loaded_file_info.object = self.get_loaded_file_info()

    def get_loaded_file_info(self) -> str:
        file_counts: dict[str, int] = self.controller.get_loaded_file_counts()
        count_str: str = ""
        for filetype in file_counts:
            count_str += f"{filetype}: {file_counts[filetype]}\n"

        return count_str

    def _set_build_buttons_status(self, active: bool, *_):
        if active:
            self.corpus_name_input.visible = True
            self.build_button.visible = True
            self.unload_all_button.disabled = False
            self.unload_selected_button.disabled = False
        else:
            self.corpus_name_input.visible = False
            self.build_button.visible = False
            self.unload_all_button.disabled = True
            self.unload_selected_button.disabled = True

    def load_as_corpus(self, *_):
        file_ls: list[FileReference] = self.file_selector.get_selector_value()
        self.view_handler.load_corpus_from_filepaths(file_ls)

    def load_as_meta(self, *_):
        file_ls: list[FileReference] = self.file_selector.get_selector_value()
        self.view_handler.load_meta_from_filepaths(file_ls)

    def unload_selected(self, *_):
        file_ls: list[FileReference] = self.file_selector.get_selector_value()
        self.controller.unload_filepaths(file_ls)
        self.view_handler.update_displays()

    def unload_all(self, *_):
        self.controller.unload_all()
        self.view_handler.update_displays()

    def build_corpus(self, *_):
        self.view_handler.build_corpus(self.corpus_name_input.value)
        self.corpus_name_input.value = ""
